/*  fgwirc - fungw based minimal irc client
    Copyright (C) 2019,2021  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page and contact: http://repo.hu/projects/fgwirc
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <libfungw/fungw.h>
#include <sockev/sockev.h>
#include <sockev/sockev_tcp4.h>
#include <sockev/filt_split.h>
#include <genht/hash.h>
#include <qparse/qparse.h>
#include "action.h"
#include "channel.h"
#include "window.h"
#include "fgwirc.h"
#include "conf.h"
#include "script.h"

#include "irc.h"

#warning TODO: the live copy of these should be in the server_t and these should be defaults only
char *irc_ident = "fgwirc";
char *irc_user  = "fgwirc";
char *irc_nick  = "fgwirc";
int irc_default_port = 6667;

vtp0_t irc_servers;
int irc_debug = 0;

void irc_init(void)
{
	conf_reg("irc_ident", FGW_STR, &irc_ident, NULL, "default user ident when connecting to a new IRC server");
	conf_reg("irc_user", FGW_STR, &irc_user, NULL, "default user name when connecting to a new IRC server");
	conf_reg("irc_nick", FGW_STR, &irc_nick, NULL, "default nickname when connecting to a new IRC server");
	conf_reg("irc_debug", FGW_INT, &irc_debug, NULL, "print all incoming lines from the IRC servers in the server status window");
	conf_reg("irc_default_port", FGW_INT, &irc_default_port, NULL, "port to be used when connecting to a new IRC server with port unspecified");
}

/* Copy and parse a nick!ident@host into fields; assumes strlen(from) < 514 */
static void parse_nick(irc_from_t *dst, const char *from)
{
	const char *si;
	char *so, *sn;

	si = from;
	if (*si == ':')
		si++;

	/* parse nick and make two copies */
	for(so = dst->copy+1, sn = dst->nick; (*si != '\0') && (*si != '!'); si++,so++,sn++) {
		*so = *si;
		*sn = *si;
	}
	*so = '\0';
	*sn = '\0';
	dst->nicklen = sn - dst->nick;

	/* parse ident */
	so = dst->ident;
	if (*si == '!') {
		for(si++; (*si != '@') && (*si != '\0'); si++,so++)
			*so = *si;
	}
	*so = '\0';
	dst->identlen = so - dst->ident;

	/* parse host */
	so = dst->host;
	if (*si == '@') {
		for(si++; (*si != '\0'); si++,so++)
			*so = *si;
	}
	*so = '\0';
	dst->hostlen = so - dst->host;
}

static void irc_parse_target(irc_server_t *server, win_t **win, const char **target)
{
	if (*target == NULL) *target = "???";
	if (**target == ':') (*target)++;
	*win = win_target2win(TARGET_CHANNEL, server->sid, *target);
	if (*win == NULL)
		*win = server->status_win;
}

void irc_win_deliver_incoming(irc_server_t *server, const irc_from_t *fr, const char *target, const char *line)
{
	win_t *win;
	char tmp2[1600];

	if (irc_is_prefix_chan(*target)) {
		/* display on channel's win */
		win = win_target2win(TARGET_CHANNEL, server->sid, target);
	}
	else {
		/* display on sender nick's win */
		win = win_target2win(TARGET_NICK, server->sid, fr->nick);
		if (win == NULL) {
			sprintf(tmp2, "%d", server->sid);
			action_vstr("on_query_alloc", tmp2, fr->nick, NULL);
			win = win_target2win(TARGET_NICK, server->sid, fr->nick);
		}
	}
	if (win == NULL) {
		/* final fallback */
		sprintf(tmp2, "*** incoming message with no target window from %s (%s@%s):", fr->nick, fr->ident, fr->host);
		win = server->status_win;
		win_append(win, tmp2);
	}
	win_append(win, line);
}

#define ARGV_OR(idx, invalid) (((idx) < argc) ? argv[idx] : invalid)
#define ARGV(idx) ARGV_OR(idx, "???")
static const char *irc_payload(const char *p) { return (*p == ':') ? p+1 : p; }

/* append user activity in server status window and in current window if it has the same sid */
#define append_win_by_sid(str) \
do { \
	win_append(server->status_win, str); \
	win = idx2win(win_curr_idx); \
	if (win->target.sid == server->sid) \
		win_append(win, str); \
} while(0)

static int irc_parse_high(irc_server_t *server, irc_from_t *fr, const char *low_cmd, int argc, char**argv)
{
	int int_cmd;
	char *tmp = fr->copy, *so;
	const char *sect;
	win_t *win;

	int_cmd = atoi(low_cmd);
	if (int_cmd == 0) {

		if (strcmp(low_cmd, "join") == 0) {
			const char *chan = ARGV(1);

			irc_parse_target(server, &win, &chan);
			sprintf(tmp, "*** %s (%s@%s) joins %s", fr->nick, fr->ident, fr->host, chan);
			win_append(win, tmp);
		}
		else if (strcmp(low_cmd, "part") == 0) {
			const char *chan = ARGV(1);
			const char *msg = ARGV(2);

			irc_parse_target(server, &win, &chan);
			sprintf(tmp, "*** %s (%s@%s) parts %s (%s)", fr->nick, fr->ident, fr->host, chan, irc_payload(msg));
			win_append(win, tmp);
		}
		else if (strcmp(low_cmd, "quit") == 0) {
			const char *msg = ARGV(1);

			sprintf(tmp, "*** %s (%s@%s) quits (%s)", fr->nick, fr->ident, fr->host, irc_payload(msg));
			append_win_by_sid(tmp);
		}
		else if (strcmp(low_cmd, "nick") == 0) {
			if (strcmp(fr->nick, server->nick) == 0) {
				sprintf(tmp, "*** You have changed nick to %s", ARGV(1));
				strncpy(server->nick, ARGV(1), sizeof(server->nick));
			}
			else
				sprintf(tmp, "*** %s has changed nick to %s", fr->nick, ARGV(1));
			append_win_by_sid(tmp);
		}
		else if (strcmp(low_cmd, "invite") == 0) {
			sprintf(tmp, "*** %s invites you to join %s", fr->nick, ARGV(2));
			win_append(server->status_win, tmp);
		}
		else if (strcmp(low_cmd, "kick") == 0) {
			const char *chan = ARGV(1);
			const char *poor = ARGV(2);
			const char *msg = ARGV(3);

			irc_parse_target(server, &win, &chan);
			sprintf(tmp, "*** %s kicks %s from %s (%s)", fr->nick, poor, chan, irc_payload(msg));
			win_append(win, tmp);
		}
		else if ((strcmp(low_cmd, "mode") == 0) || (strcmp(low_cmd, "topic") == 0)) {
			const char *chan = ARGV(1);
			char *end = tmp;
			int n;

			irc_parse_target(server, &win, &chan);
			end += sprintf(tmp, "*** %s (%s@%s) sets %s on %s: ", fr->nick, fr->ident, fr->host, low_cmd, chan);
			for(n = 2; n < argc; n++)
				end += sprintf(end, " %s", irc_payload(argv[n]));
			win_append(win, tmp);
		}
		else if ((strcmp(low_cmd, "privmsg") == 0) || (strcmp(low_cmd, "notice") == 0) || (strncmp(low_cmd, "ctcp_", 5) == 0)) {
			switch(low_cmd[0]) {
				case 'p': tmp[0] = '<'; break;
				case 'n': tmp[0] = '-'; break;
				case 'c': tmp[0] = '^'; break;
				default:  tmp[0] = '?'; break;
			}
			so = fr->copy + fr->nicklen+1;
			switch(low_cmd[0]) {
				case 'p': *so = '>'; break;
				case 'n': *so = '-'; break;
				case 'c': *so = '^'; break;
				default:  *so = '?'; break;
			}
			so++;
			*so = ' '; so++;
			strcpy(so, ARGV(2));
			irc_win_deliver_incoming(server, fr, ARGV(1), tmp);
		}
	}
	else  {
		switch(int_cmd) {
			case 1: case 2: case 3: case 5: break; /* initial conn ignores */
			case 4:
				channel_join_pending_by_sid(server->sock, server->sid);
				break;

			/* whois */
			case 301: sect = "is away"; goto whois;
			case 311: sect = "is user"; goto whois;
			case 312: sect = "using server"; goto whois;
			case 313: sect = ""; goto whois;
			case 317: sect = "idles"; goto whois;
			case 318: sect = ""; goto whois;
			case 319: sect = "is on"; goto whois;

			/* away state change feedback */
			case 305: sect = ""; goto whois; 
			case 306: sect = ""; goto whois; 

			/* who */
			case 352: sect = ""; goto whois;
			case 315: sect = ""; goto whois;

			/* whowas */
			case 314: sect = ""; goto whois;
			case 369: sect = ""; goto whois;

			case 324: /* chan mode */
				{
					const char *chan  = irc_payload(ARGV(2)), *mode = irc_payload(ARGV(3));
					win = win_target2win(TARGET_CHANNEL, server->sid, chan);
					if (win == NULL) win = server->status_win;
					sprintf(tmp, "*** Channel mode %s: %s", chan, mode);
					win_append(win, tmp);
				}
				break;

			case 332: case 331: /* chan topic */
				{
					const char *chan  = irc_payload(ARGV(2)), *topic = irc_payload(ARGV(3));
					win = win_target2win(TARGET_CHANNEL, server->sid, chan);
					if (win == NULL) win = server->status_win;
					sprintf(tmp, "*** Topic %s: %s", chan, topic);
					win_append(win, tmp);
				}
				break;
			case 333: /* chan topic set nick & date */
				{
					const char *chan = irc_payload(ARGV(2)), *setnick = irc_payload(ARGV(3)), *whens = irc_payload(ARGV(4));
					long when = strtol(whens, NULL, 10), len;
					time_t t = when;
					win = win_target2win(TARGET_CHANNEL, server->sid, chan);

					if (win == NULL) win = server->status_win;
					len = sprintf(tmp, "*** topic on %s set by %s at %s", chan, setnick, ctime(&t));
					tmp[len-1] = '\0';
					win_append(win, tmp);
				}
				break;
			case 353: /* chan names */
				win = win_target2win(TARGET_CHANNEL, server->sid, ARGV(3));
				if (win != NULL) {
					if (win->target.state == TGTST_INPROG) {
						sprintf(tmp, "*** Joined %d/%s", win->target.sid, win->target.name);
						win_append(win, tmp);
						win_set_target_state(win, TGTST_NORMAL);
						win_set_name(win, win->target.name);
					}
				}
				else
					win = server->status_win;
				sprintf(tmp, "Names: %s: %s", ARGV(3), ARGV(4));
				win_append(win, tmp);
				break;

			case 341: /* invite feedback */
				sprintf(tmp, "invited %s to %s", ARGV(2), ARGV(3));
				win_append(server->status_win, tmp);
				break;

			/* links */
			case 364:
				sprintf(tmp, "[LNK] %s %s :%s", ARGV(2), ARGV(3), ARGV(4));
				win_append(server->status_win, tmp);
				break;
			case 365:
				sprintf(tmp, "[LNK] %s from %s", ARGV(3), fr->nick);
				win_append(server->status_win, tmp);
				break;

			case 366: break; /* end of name list */

			case 372:
			case 375:
			case 376:
				sprintf(tmp, "[MOTD] %s", ARGV(2));
				win_append(server->status_win, tmp);
				break;

			case 433: /* nick already in use */
				sprintf(tmp, "*** ERROR: (422) Nickname %s is already in use at %s", ARGV(2), fr->nick);
				win_append(server->status_win, tmp);
				break;

			default:
				sprintf(tmp, "[%d] '%s' '%s' '%s'", int_cmd, ARGV_OR(2, ""), ARGV_OR(3, ""), ARGV_OR(4, ""));
				win_append(server->status_win, tmp);
		}
	}

	return 0;

	whois:;
	{
		const char *nick = ARGV(2);
		char *end = tmp;
		int n;

		end += sprintf(end, "*** %s %s", nick, sect);
		for(n = 3; n < argc; n++)
			end += sprintf(end, " %s", irc_payload(argv[n]));

		win = idx2win(win_curr_idx);
		if (win->target.sid == server->sid)
			win_append(win, tmp);
		else
			win_append(server->status_win, tmp);
	}
	return 0;
}

static int irc_parse_low(sockev_t sock, irc_server_t *server, char *line, int len)
{
	irc_from_t from;
	char *end, *cmd, *from_str = line, low_cmd[512], *si, *so;
	char **argv;
	int argc;

	if (irc_debug)
		win_append(server->status_win, line);

	if (len > 512) { /* check size early on so buffers can use static size */
		win_append(server->status_win, "*** IRC protocol voilation: message too long:");
		win_append(server->status_win, line);
		goto error;
	}

	if (strncmp(line, "PING", 4) == 0) {
		from_str[1] = 'O';
		sockev_write(sock, line, len);
		sockev_write(sock, "\r\n", 2);
		return 0;
	}

	cmd = strpbrk(line, " \t\r\n");
	if (cmd == NULL)
		return 0;
	*cmd = '\0';
	cmd++;

	/* from, composite string variant */
	end = strpbrk(from_str, " \t\r\n");
	if (end != NULL) {
		if (*from_str == ':')
			from_str++;
		*end = 0;
	}

	argc = qparse2(cmd, &argv, QPARSE_MULTISEP | QPARSE_TERM_NEWLINE | QPARSE_COLON_LAST);

	/* convert argv[0] to lowercase for simpler processing; leave 8 bytes free
	   for prefix (optimization) */
	for(si = argv[0], so=low_cmd+8; *si != '\0'; si++,so++)
		*so = tolower(*si);
	*so = '\0';

	parse_nick(&from, from_str);

	if ((argc > 2) && (argv[2][0] == '\001')) {
		char *s, *last = argv[argc-1];
		int lastlen = strlen(last);

		/* remove trailing \001 */
		if (lastlen > 0) {
			last += (lastlen - 1);
			if (*last == '\001')
				*last = '\0';
		}

		/* remove leading \001 */
		for(s = argv[2]+1; *s != '\0'; s++)
			s[-1] = *s;
		s[-1] = '\0';

		/* replace command */
		if (strcmp(low_cmd+8, "privmsg") == 0)
			strcpy(low_cmd+8, "ctcp_req");
		else if (strcmp(low_cmd+8, "notice") == 0)
			strcpy(low_cmd+8, "ctcp_resp");
	}

	if (script_parse_high(SCRIPT_PRE, server, &from, low_cmd+8, argc, argv) != ACTION_HOOK_GO_ON)
		goto error;
	if (script_parse_high(SCRIPT_ON, server, &from, low_cmd+8, argc, argv) != ACTION_HOOK_GO_ON)
		goto error;
	if (irc_parse_high(server, &from, low_cmd+8, argc, argv) < 0)
		goto error;
	script_parse_high(SCRIPT_POST, server, &from, low_cmd+8, argc, argv);

	error:;
	qparse_free(argc, &argv);

	return 0;
}

static void ev_irc_open(sockev_t id)
{
	irc_server_t *server = sockev_get_userdata(id);

	win_append(server->status_win, "*** TCP connection made");
	sockev_printf(id, "user %s b c :%s\n", irc_ident, irc_user);
	sockev_printf(id, "nick %s\n", server->want_nick);
}

static void ev_irc_close(sockev_t id)
{
	irc_server_t *server = sockev_get_userdata(id);
	win_append(server->status_win, "*** TCP connection closed");
	channel_disconnected_by_sid(server->sid);
}

static void ev_irc_read(sockev_t id, char *buff, int len)
{
	irc_server_t *server = sockev_get_userdata(id);
	irc_parse_low(id, server, buff, len);
}

static sockev_hooks_t ev_irc = {ev_irc_open, ev_irc_close, NULL, NULL, NULL, ev_irc_read, NULL};
static sockev_filters_t flt_linesplit[] = {
	{sockev_filter_split, "\r\n"},
	{NULL, NULL}
};

void irc_reconnect(int sid)
{
	char tmp[1024];
	void **v = vtp0_get(&irc_servers, sid, 0);
	irc_server_t *server;

	if ((v == NULL) || (*v == NULL)) {
		sprintf(tmp, "Can't connect to server id %d: not such server", sid);
		win_append(wstatus, tmp);
		return;
	}
	server = *v;

	if (server->sock != SOCKEV_INVALID) {
		sprintf(tmp, "Can't connect server %d/%s:%d becasue server id already has a network connection\n", sid, server->hostname, server->port);
		win_append(server->status_win, tmp);
	}

	sprintf(tmp, "Connecting server %d/%s:%d\n", sid, server->hostname, server->port);
	win_append(server->status_win, tmp);


	server->sock = sockev_new(&ev_irc, 0, flt_linesplit);
	sockev_set_userdata(server->sock, server);
	if (sockev_tcp4_connect(server->sock, server->hostname, server->port) == SOCKEV_INVALID) {
		sprintf(tmp, "Failed to connect server server %d/%s:%d\n", sid, server->hostname, server->port);
		win_append(server->status_win, tmp);
		return;
	}
}

int irc_connect_new(win_t *win, const char *hostname, int port, int sid)
{
	irc_server_t *server;
	char tmp[600];

	if (strlen(hostname) > 512)
		return -1;

	if ((sid >= 0) && (sid < irc_servers.used))  {
		irc_disconnect(sid);
		server = irc_servers.array[sid];
		free(server->hostname);
		server->hostname = NULL;
	}
	else {
		if (sid < 0)
			sid = irc_servers.used;
		server = calloc(sizeof(irc_server_t), 1);
	}


	sprintf(tmp, "{server %s:%d}", hostname, port);
	win_set_target(win, TARGET_SERVER_STATUS, sid, tmp, NULL);

	server->status_win = win;
	server->hostname = fgw_strdup(hostname);
	server->port = port;
	server->status_win = win;
	server->state = IRCS_OFFLINE;
	server->sock = SOCKEV_INVALID;
	server->sid = sid;

	strncpy(server->nick, irc_nick, sizeof(server->want_nick));
	strncpy(server->want_nick, irc_nick, sizeof(server->want_nick));

	vtp0_set(&irc_servers, sid, server);

	irc_reconnect(sid);

	return sid;
}

void irc_disconnect(int sid)
{
	char tmp[1024];
	void **v = vtp0_get(&irc_servers, sid, 0);
	irc_server_t *server;

	if ((v == NULL) || (*v == NULL)) {
		sprintf(tmp, "Can't disconnect from server id %d: not such server", sid);
		win_append(wstatus, tmp);
		return;
	}
	server = *v;

	sprintf(tmp, "Disconnecting from server %d/%s:%d\n", sid, server->hostname, server->port);
	win_append(server->status_win, tmp);
	
}


sockev_t irc_server_sock(int sid)
{
	void **v = vtp0_get(&irc_servers, sid, 0);
	irc_server_t *server;

	if ((v == NULL) || (*v == NULL))
		return SOCKEV_INVALID;
	server = *v;
	return server->sock;
}

static irc_server_t *irc_get_win_server_(win_t *win)
{
	if ((win->target.sid < 0) || (win->target.sid >= irc_servers.used))
		return NULL;
	return irc_servers.array[win->target.sid];
}

irc_server_t *irc_get_win_server(win_t *win) { return irc_get_win_server_(win); }

static const char *irc_get_win_my_nick_(win_t *win, const char *fallback)
{
	irc_server_t *server = irc_get_win_server_(win);
	if ((server == NULL) || (server->nick == NULL))
		return fallback;
	return server->nick;
}

const char *irc_get_win_my_nick(win_t *win, const char *fallback) { return irc_get_win_my_nick_(win, fallback); }


static void echo(win_t *win, const char *str, int succ)
{
	char tmp[1024]; /* safe: str is size limited on edit.c and nick is finite too */
	const char *mynick = irc_get_win_my_nick_(win, irc_nick);

	sprintf(tmp, "%s<%s> %s", succ ? "" : "!", mynick, str);
	win_append(win, tmp);
}

int irc_send_win_msg(win_t *win, const char *str)
{
	sockev_t sock = irc_server_sock(win->target.sid);
	int succ = 1;

	if (sock == SOCKEV_INVALID) {
		win_append(win, "Can not message: server not connected");
		succ = 0;
	}
	else if (*str != '\0')
		sockev_printf(sock, "PRIVMSG %s :%s\r\n", win->target.name, str);

	echo(win, str, succ);
	return !succ;
}

int irc_send_win_raw(win_t *win, const char *str)
{
	sockev_t sock = irc_server_sock(win->target.sid);
	int succ = 1;

	if (sock == SOCKEV_INVALID) {
		win_append(win, "Can not send command: server not connected");
		succ = 0;
	}
	else if (*str != '\0')
		sockev_printf(sock, "%s\r\n", str);

	echo(win, str, succ);
	return !succ;
}

int irc_send_all_raw(const char *str)
{
	int n;
	for(n = 0; n < irc_servers.used; n++) {
		irc_server_t *server = irc_servers.array[n];
		if ((server != NULL) && (server->sock != SOCKEV_INVALID))
			sockev_printf(server->sock, "%s\r\n", str);
	}
	return 0;
}

int irc_lookup_server(const char *name)
{
	int n;

	if (isdigit(*name)) {
		char *end;
		int sid = strtol(name, &end, 10);
		if ((*end != '\0') || (sid < 0))
			return -1;
		return sid;
	}

	for(n = 0; n < irc_servers.used; n++) {
		irc_server_t *server = irc_servers.array[n];
		if (genht_strcasecmp(server->hostname, name) == 0)
			return n;
	}

	return -1;
}
