/*  fgwirc - fungw based minimal irc client
    Copyright (C) 2019,2021  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page and contact: http://repo.hu/projects/fgwirc
*/

#include <stdio.h>
#include <dlfcn.h>
#include <genvector/gds_char.h>
#include <genht/htsp.h>
#include <genht/hash.h>

#include "action.h"

#include "script.h"

char *fungw_libdir = FUNGW_PREFIX "/lib";

htsp_t scripts;

static const char *script_fn(win_t *win, const char *fn, char *buff, size_t buffsize)
{
	if (*fn == '~') {
		const char *home = getenv("HOME");
		if (strlen(home) + strlen(fn) > buffsize-4) {
			win_append(win, "Failed to load the script: path too long (~)");
			return NULL;
		}
		sprintf(buff, "%s%s", home, fn+1);
		fn = buff;
	}
	return fn;
}

static win_t *ase_win;
static void load_async_error(fgw_obj_t *obj, const char *msg)
{
	win_append_strs(ase_win, "Runtime error on object '", obj->name, "': ", msg, NULL);
}

int script_load(win_t *win, const char *id, const char *fn, const char *lang)
{
	size_t langlen, fnlen;
	char tmp[1024];
	void *eng;
	fgw_obj_t *obj;
	void (*init)(void);
	script_t *sc;

	if ((strcmp(lang, "act") == 0) || (strcmp(lang, "action") == 0)) {
		fn = script_fn(win, fn, tmp, sizeof(tmp));
		if (fn == NULL) return -1;
		action_load_file(fn);
		return 0;
	}

	langlen = strlen(lang);
	fnlen = strlen(fn);

	fgw.async_error = load_async_error;
	ase_win = win;

	if (strcmp(fn+fnlen-3, ".so") == 0) {
		eng = dlopen(fn, RTLD_GLOBAL | RTLD_NOW);
		if (eng == NULL) {
			win_append(win, "Failed to load the .so plugin:");
			win_append(win, dlerror());
			return -1;
		}
		goto loaded;
	}

#warning TODO: move this to fungw
	if (strcmp(lang, "fbas") == 0) lang = "fawk";
	else if (strcmp(lang, "fpas") == 0) lang = "fawk";

	if (langlen + strlen(fungw_libdir) < sizeof(tmp)-10) {
		sprintf(tmp, "%s/puplug/fungw_%s.so", fungw_libdir, lang);
		eng = dlopen(tmp, RTLD_GLOBAL | RTLD_NOW);
		if (eng == NULL) {
			win_append(win, "Failed to load the script lang engine:");
			win_append(win, dlerror());
			return -1;
		}
		loaded:;
		sprintf(tmp, "pplg_init_fungw_%s", lang);
		init = dlsym(eng, tmp);
		if (init == NULL) {
			win_append(win, "Failed to load the script lang engine: can't find the puplug init call");
			return -1;
		}
		init();
	}


	fn = script_fn(win, fn, tmp, sizeof(tmp));
	if (fn == NULL) return -1;
	obj = fgw_obj_new(&fgw, id, lang, fn, NULL);
	if (obj == NULL) {
		win_append(win, "Failed to load the script:");
		win_append(win, fn);
		return -1;
	}

	sc = malloc(sizeof(script_t));
	sc->obj = obj;
	sc->fn = fgw_strdup(fn);
	sc->lang = fgw_strdup(lang);
	htsp_set(&scripts, obj->name, sc);

	script_clean_hooks();
	return 0;
}

int script_unload(win_t *win, const char *id)
{
	script_t *sc = htsp_pop(&scripts, id);

	if (sc == NULL) {
		win_append(win, "No such script id");
		return -1;
	}

	fgw_obj_unreg(&fgw, sc->obj);

	free(sc->fn);
	free(sc->lang);
	free(sc);
	script_clean_hooks();

	return 0;
}

void script_init(void)
{
	htsp_init(&scripts, strhash, strkeyeq);
}

void script_uninit(void)
{
	htsp_entry_t *e;
	for(e = htsp_first(&scripts); e != NULL; e = htsp_next(&scripts, e)) {
		script_t *sc = e->value;
		fgw_obj_unreg(&fgw, sc->obj);
		free(sc->fn);
		free(sc->lang);
		free(sc);
	}
	htsp_uninit(&scripts);
}

action_hook_res_t script_parse_high(script_hook_type_t type, irc_server_t *server, const irc_from_t *from, char *low_cmd, int argc, char **argv)
{
	action_details_t actd;
	const char *stype;
	int stype_len;

	switch(type) {
		case SCRIPT_PRE:  stype = "pre_"; stype_len = 4; break;
		case SCRIPT_POST: stype = "post_"; stype_len = 5; break;
		case SCRIPT_ON:   stype = "on_";   stype_len = 3; break;
		default: return ACTION_HOOK_GO_ON;
	}

	action_details_reset(&actd);

	actd.signature = actd_signature;
	actd.server = server;
	actd.from = from;
	actd.argc = argc;
	actd.argv = argv;
	actd.res = ACTION_HOOK_GO_ON;


	/* NOTE: there are 8 bytes reserved before low_cmd for prefixing! */
	memcpy(low_cmd - stype_len, stype, stype_len);

	fgw_uvcall_all(&fgw, &actd, low_cmd-stype_len, FGW_STR, from->nick, FGW_INT, argc, 0);
	return actd.res;
}

void script_clean_hooks(void)
{
	static vtp0_t cache;
	htsp_entry_t *e;
	int n;

	cache.used = 0;
	for(e = htsp_first(&fgw.func_tbl); e != NULL; e = htsp_next(&fgw.func_tbl, e)) {
		char *s = e->key;
		if ((s[0] == 'o') && (s[1] == 'n') && (s[2] == '_')) {}
		else if ((s[0] == 'p') && (s[1] == 'r') && (s[2] == 'e') && (s[3] == '_')) {}
		else if ((s[0] == 'p') && (s[1] == 'o') && (s[2] == 's') && (s[3] == 't') && (s[4] == '_')) {}
		else continue;
		vtp0_append(&cache, e->key);
	}

	for(n = 0; n < cache.used; n++)
		htsp_pop(&fgw.func_tbl, cache.array[n]);
}
